@@ -28,6 +28,7 @@
 #include "librpc/gen_ndr/security.h"
 #include "dom_sid.h"
 #include "lib/util/smb_strtox.h"
+#include "lib/util/tsort.h"
 
 /*****************************************************************
  Compare the auth portion of two sids.
@@ -38,19 +39,27 @@ int dom_sid_compare_auth(const struct dom_sid *sid1,
 {
 	int i;
 
-	if (sid1 == sid2)
+	if (sid1 == sid2) {
 		return 0;
-	if (!sid1)
+	}
+
+	if (sid1 == NULL) {
 		return -1;
-	if (!sid2)
+	}
+
+	if (sid2 == NULL) {
 		return 1;
+	}
 
-	if (sid1->sid_rev_num != sid2->sid_rev_num)
-		return sid1->sid_rev_num - sid2->sid_rev_num;
+	if (sid1->sid_rev_num != sid2->sid_rev_num) {
+		return NUMERIC_CMP(sid1->sid_rev_num, sid2->sid_rev_num);
+	}
 
-	for (i = 0; i < 6; i++)
-		if (sid1->id_auth[i] != sid2->id_auth[i])
-			return sid1->id_auth[i] - sid2->id_auth[i];
+	for (i = 0; i < 6; i++) {
+		if (sid1->id_auth[i] != sid2->id_auth[i]) {
+			return NUMERIC_CMP(sid1->id_auth[i], sid2->id_auth[i]);
+		}
+	}
 
 	return 0;
 }
@@ -63,17 +72,22 @@ int dom_sid_compare(const struct dom_sid *sid1, const struct dom_sid *sid2)
 {
 	int i;
 
-	if (sid1 == sid2)
+	if (sid1 == sid2) {
 		return 0;
-	if (!sid1)
+	}
+
+	if (sid1 == NULL) {
 		return -1;
-	if (!sid2)
+	}
+
+	if (sid2 == NULL) {
 		return 1;
+	}
 
 	/* Compare most likely different rids, first: i.e start at end */
-	if (sid1->num_auths != sid2->num_auths)
-		return sid1->num_auths - sid2->num_auths;
-
+	if (sid1->num_auths != sid2->num_auths) {
+		return NUMERIC_CMP(sid1->num_auths, sid2->num_auths);
+	}
 	for (i = sid1->num_auths-1; i >= 0; --i) {
 		if (sid1->sub_auths[i] < sid2->sub_auths[i]) {
 			return -1;
@@ -360,14 +374,37 @@ NTSTATUS dom_sid_split_rid(TALLOC_CTX *mem_ctx, const struct dom_sid *sid,
 	return NT_STATUS_OK;
 }
 
+/*
+  return true if the 2nd sid contains or matches the prefix_sid
+*/
+bool dom_sid_match_prefix(const struct dom_sid *prefix_sid,
+			  const struct dom_sid *sid)
+{
+	int i;
+
+	if (!prefix_sid || !sid) {
+		return false;
+	}
+
+	if (prefix_sid->num_auths > sid->num_auths) {
+		return false;
+	}
+
+	for (i = prefix_sid->num_auths-1; i >= 0; --i) {
+		if (prefix_sid->sub_auths[i] != sid->sub_auths[i]) {
+			return false;
+		}
+	}
+
+	return dom_sid_compare_auth(prefix_sid, sid) == 0;
+}
+
 /*
   return true if the 2nd sid is in the domain given by the first sid
 */
 bool dom_sid_in_domain(const struct dom_sid *domain_sid,
 		       const struct dom_sid *sid)
 {
-	int i;
-
 	if (!domain_sid || !sid) {
 		return false;
 	}
@@ -380,13 +417,7 @@ bool dom_sid_in_domain(const struct dom_sid *domain_sid,
 		return false;
 	}
 
-	for (i = domain_sid->num_auths-1; i >= 0; --i) {
-		if (domain_sid->sub_auths[i] != sid->sub_auths[i]) {
-			return false;
-		}
-	}
-
-	return dom_sid_compare_auth(domain_sid, sid) == 0;
+	return dom_sid_match_prefix(domain_sid, sid);
 }
 
 bool dom_sid_has_account_domain(const struct dom_sid *sid)
