@@ -27,10 +27,13 @@
 #include "../libcli/smb/smb_common.h"
 #include "../librpc/gen_ndr/ndr_security.h"
 #include "../librpc/gen_ndr/ndr_smb2_lease_struct.h"
+#include "../librpc/gen_ndr/ndr_smb3posix.h"
 #include "../lib/util/tevent_ntstatus.h"
 #include "messages.h"
 #include "lib/util_ea.h"
 #include "source3/passdb/lookup_sid.h"
+#include "source3/modules/util_reparse.h"
+#include "libcli/smb/reparse.h"
 
 #undef DBGC_CLASS
 #define DBGC_CLASS DBGC_SMB2
@@ -118,7 +121,7 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 			uint32_t in_file_attributes,
 			uint32_t in_share_access,
 			uint32_t in_create_disposition,
-			uint32_t in_create_options,
+			uint32_t _in_create_options,
 			const char *in_name,
 			struct smb2_create_blobs in_context_blobs);
 static NTSTATUS smbd_smb2_create_recv(struct tevent_req *req,
@@ -134,7 +137,8 @@ static NTSTATUS smbd_smb2_create_recv(struct tevent_req *req,
 			uint32_t *out_file_attributes,
 			uint64_t *out_file_id_persistent,
 			uint64_t *out_file_id_volatile,
-			struct smb2_create_blobs *out_context_blobs);
+			struct smb2_create_blobs *out_context_blobs,
+			struct reparse_data_buffer **symlink_reparse);
 
 static void smbd_smb2_request_create_done(struct tevent_req *tsubreq);
 NTSTATUS smbd_smb2_request_process_create(struct smbd_smb2_request *smb2req)
@@ -268,6 +272,26 @@ NTSTATUS smbd_smb2_request_process_create(struct smbd_smb2_request *smb2req)
 		return smbd_smb2_request_error(smb2req, status);
 	}
 
+	if (CHECK_DEBUGLVL(DBGLVL_DEBUG)) {
+		char *str = talloc_asprintf(
+			talloc_tos(),
+			"\nGot %"PRIu32" create blobs\n",
+			in_context_blobs.num_blobs);
+		uint32_t i;
+
+		for (i=0; i<in_context_blobs.num_blobs; i++) {
+			struct smb2_create_blob *b =
+				&in_context_blobs.blobs[i];
+			talloc_asprintf_addbuf(&str, "[%"PRIu32"]\n", i);
+			dump_data_addbuf(
+				(uint8_t *)b->tag, strlen(b->tag), &str);
+			dump_data_addbuf(
+				b->data.data, b->data.length, &str);
+		}
+		DBG_DEBUG("%s", str);
+		TALLOC_FREE(str);
+	}
+
 	tsubreq = smbd_smb2_create_send(smb2req,
 				       smb2req->sconn->ev_ctx,
 				       smb2req,
@@ -295,6 +319,122 @@ static uint64_t get_mid_from_smb2req(struct smbd_smb2_request *smb2req)
 	return BVAL(reqhdr, SMB2_HDR_MESSAGE_ID);
 }
 
+/*
+ * [MS-SMB2] 2.2.2.1 SMB2 ERROR Context Response
+ */
+static bool smbd_smb2_create_symlink_error_context_response(
+	TALLOC_CTX *mem_ctx,
+	struct reparse_data_buffer *symlink_reparse,
+	uint8_t **_resp,
+	size_t *_resplen)
+{
+	ssize_t symlink_buffer_len;
+	size_t symlink_error_response_len, error_context_response_len;
+	uint8_t *resp = NULL;
+
+	SMB_ASSERT(symlink_reparse->tag == IO_REPARSE_TAG_SYMLINK);
+
+	symlink_buffer_len = reparse_data_buffer_marshall(symlink_reparse,
+							  NULL,
+							  0);
+	if (symlink_buffer_len < 0) {
+		DBG_DEBUG("reparse_data_buffer_marshall() failed\n");
+		goto fail;
+	}
+
+	/* 2.2.2.2.1 Symbolic Link Error Response */
+	symlink_error_response_len = symlink_buffer_len + 8;
+	if (symlink_error_response_len < (size_t)symlink_buffer_len) {
+		goto fail;
+	}
+
+	/* 2.2.2.1 SMB2 ERROR Context Response */
+	error_context_response_len = symlink_error_response_len + 8;
+	if (error_context_response_len < symlink_error_response_len) {
+		goto fail;
+	}
+
+	resp = talloc_array(mem_ctx, uint8_t, error_context_response_len);
+	if (resp == NULL) {
+		goto fail;
+	}
+	PUSH_LE_U32(resp, 0, symlink_error_response_len); /* ErrorDataLength */
+	PUSH_LE_U32(resp, 4, 0);			  /* ErrorId */
+	PUSH_LE_U32(resp,
+		    8,
+		    symlink_error_response_len - 4); /* SymLinkLength */
+	PUSH_LE_U32(resp, 12, 0x4C4D5953);
+
+	reparse_data_buffer_marshall(symlink_reparse,
+				     resp + 16,
+				     symlink_buffer_len);
+
+	*_resp = resp;
+	*_resplen = error_context_response_len;
+	return true;
+fail:
+	TALLOC_FREE(resp);
+	return false;
+}
+
+static NTSTATUS smbd_smb2_create_error(
+	struct smbd_smb2_request *smb2req,
+	NTSTATUS status,
+	struct reparse_data_buffer *symlink_reparse)
+{
+	struct smbXsrv_connection *xconn = smb2req->xconn;
+	NTSTATUS error;
+	DATA_BLOB resp = {
+		.data = NULL,
+	};
+	bool ok;
+
+	if (!NT_STATUS_EQUAL(status, NT_STATUS_STOPPED_ON_SYMLINK)) {
+		error = smbd_smb2_request_error(smb2req, status);
+		return error;
+	}
+
+	ok = smbd_smb2_create_symlink_error_context_response(smb2req,
+							     symlink_reparse,
+							     &resp.data,
+							     &resp.length);
+	if (!ok) {
+		error = smbd_smb2_request_error(smb2req, NT_STATUS_NO_MEMORY);
+		return error;
+	}
+
+	if (xconn->protocol < PROTOCOL_SMB3_11) {
+		/*
+		 * smbd_smb2_create_symlink_error_context_response()
+		 * has created a smb3.11 SMB2 ERROR Context from
+		 * [MS-SMB2] 2.2.2.1. This has an 8-byte header before
+		 * the symlink response.
+		 */
+		DATA_BLOB symlink_error = {
+			.data = resp.data + 8,
+			.length = resp.length - 8,
+		};
+
+		SMB_ASSERT(resp.length >= 8);
+
+		error = smbd_smb2_request_error_ex(
+			smb2req,
+			NT_STATUS_STOPPED_ON_SYMLINK,
+			0,
+			&symlink_error,
+			__location__);
+	} else {
+		error = smbd_smb2_request_error_ex(
+			smb2req,
+			NT_STATUS_STOPPED_ON_SYMLINK,
+			1,
+			&resp,
+			__location__);
+	}
+
+	return error;
+}
+
 static void smbd_smb2_request_create_done(struct tevent_req *tsubreq)
 {
 	struct smbd_smb2_request *smb2req = tevent_req_callback_data(tsubreq,
@@ -316,6 +456,7 @@ static void smbd_smb2_request_create_done(struct tevent_req *tsubreq)
 	struct smb2_create_blobs out_context_blobs;
 	DATA_BLOB out_context_buffer;
 	uint16_t out_context_buffer_offset = 0;
+	struct reparse_data_buffer *symlink_reparse = NULL;
 	NTSTATUS status;
 	NTSTATUS error; /* transport error */
 
@@ -332,12 +473,15 @@ static void smbd_smb2_request_create_done(struct tevent_req *tsubreq)
 				       &out_file_attributes,
 				       &out_file_id_persistent,
 				       &out_file_id_volatile,
-				       &out_context_blobs);
+				       &out_context_blobs,
+				       &symlink_reparse);
 	if (!NT_STATUS_IS_OK(status)) {
 		if (smbd_smb2_is_compound(smb2req)) {
 			smb2req->compound_create_err = status;
 		}
-		error = smbd_smb2_request_error(smb2req, status);
+		error = smbd_smb2_create_error(smb2req,
+					       status,
+					       symlink_reparse);
 		if (!NT_STATUS_IS_OK(error)) {
 			smbd_server_connection_terminate(smb2req->xconn,
 							 nt_errstr(error));
@@ -432,6 +576,7 @@ static NTSTATUS smbd_smb2_create_durable_lease_check(struct smb_request *smb1req
 	NTTIME twrp = fsp->fsp_name->twrp;
 	NTSTATUS status;
 	bool is_dfs = (smb1req->flags2 & FLAGS2_DFS_PATHNAMES);
+	bool is_posix = (fsp->fsp_name->flags & SMB_FILENAME_POSIX_PATH);
 
 	if (lease_ptr == NULL) {
 		if (fsp->oplock_type != LEASE_OPLOCK) {
@@ -455,19 +600,49 @@ static NTSTATUS smbd_smb2_create_durable_lease_check(struct smb_request *smb1req
 		return NT_STATUS_OBJECT_NAME_NOT_FOUND;
 	}
 
+	if (is_dfs) {
+		const char *non_dfs_requested_filename = NULL;
+		/*
+		 * With a DFS flag set, remove any DFS prefix
+		 * before further processing.
+		 */
+		status = smb2_strip_dfs_path(requested_filename,
+					     &non_dfs_requested_filename);
+		if (!NT_STATUS_IS_OK(status)) {
+			return status;
+		}
+		/*
+		 * TODO: Note for dealing with reparse point errors.
+		 * We will need to remember and store the number of characters
+		 * we have removed here, which is
+		 * (requested_filename - non_dfs_requested_filename)
+		 * in order to correctly report how many characters we
+		 * have removed before hitting the reparse point.
+		 * This will be a patch needed once we properly
+		 * deal with reparse points later.
+		 */
+		requested_filename = non_dfs_requested_filename;
+		/*
+		 * Now we're no longer dealing with a DFS path, so
+		 * remove the flag.
+		 */
+		smb1req->flags2 &= ~FLAGS2_DFS_PATHNAMES;
+		is_dfs = false;
+	}
+
 	filename = talloc_strdup(talloc_tos(), requested_filename);
 	if (filename == NULL) {
 		return NT_STATUS_NO_MEMORY;
 	}
 
 	/* This also converts '\' to '/' */
-	status = check_path_syntax_smb2(filename, is_dfs);
+	status = check_path_syntax(filename, is_posix);
 	if (!NT_STATUS_IS_OK(status)) {
 		TALLOC_FREE(filename);
 		return status;
 	}
 
-	ucf_flags = filename_create_ucf_flags(smb1req, FILE_OPEN);
+	ucf_flags = filename_create_ucf_flags(smb1req, FILE_OPEN, 0);
 	status = filename_convert_dirfsp(talloc_tos(),
 					 fsp->conn,
 					 filename,
@@ -507,8 +682,10 @@ struct smbd_smb2_create_state {
 	struct deferred_open_record *open_rec;
 	files_struct *result;
 	bool replay_operation;
+	bool replay_reconnect;
 	uint8_t in_oplock_level;
 	uint32_t in_create_disposition;
+	uint32_t in_create_options;
 	int requested_oplock_level;
 	int info;
 	char *fname;
@@ -558,6 +735,9 @@ struct smbd_smb2_create_state {
 	uint64_t out_file_id_persistent;
 	uint64_t out_file_id_volatile;
 	struct smb2_create_blobs *out_context_blobs;
+
+	/* symlink error data */
+	struct reparse_data_buffer *symlink_err;
 };
 
 static void smbd_smb2_create_purge_replay_cache(struct tevent_req *req,
@@ -683,7 +863,9 @@ static NTSTATUS smbd_smb2_create_fetch_create_ctx(
 		state->svhdx = smb2_create_blob_find(
 			in_context_blobs, SVHDX_OPEN_DEVICE_CONTEXT);
 	}
-	if (xconn->smb2.server.posix_extensions_negotiated) {
+	if (xconn->smb2.server.posix_extensions_negotiated &&
+	    lp_smb3_unix_extensions(SNUM(state->smb1req->conn)))
+	{
 		/*
 		 * Negprot only allowed this for proto>=3.11
 		 */
@@ -691,6 +873,12 @@ static NTSTATUS smbd_smb2_create_fetch_create_ctx(
 
 		state->posx = smb2_create_blob_find(
 			in_context_blobs, SMB2_CREATE_TAG_POSIX);
+		/*
+		 * Setting the bool below will cause
+		 * ucf_flags_from_smb_request() to
+		 * return UCF_POSIX_PATHNAMES in ucf_flags.
+		 */
+		state->smb1req->posix_pathnames = (state->posx != NULL);
 	}
 
 	return NT_STATUS_OK;
@@ -709,7 +897,7 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 			uint32_t in_file_attributes,
 			uint32_t in_share_access,
 			uint32_t in_create_disposition,
-			uint32_t in_create_options,
+			uint32_t _in_create_options,
 			const char *in_name,
 			struct smb2_create_blobs in_context_blobs)
 {
@@ -721,6 +909,7 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 	struct smb_filename *smb_fname = NULL;
 	uint32_t ucf_flags;
 	bool is_dfs = false;
+	bool is_posix = false;
 
 	req = tevent_req_create(mem_ctx, &state,
 				struct smbd_smb2_create_state);
@@ -732,9 +921,10 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 		.smb2req = smb2req,
 		.in_oplock_level = in_oplock_level,
 		.in_create_disposition = in_create_disposition,
+		.in_create_options = _in_create_options,
 	};
 
-	smb1req = smbd_smb2_fake_smb_request(smb2req);
+	smb1req = smbd_smb2_fake_smb_request(smb2req, NULL);
 	if (tevent_req_nomem(smb1req, req)) {
 		return tevent_req_post(req, state->ev);
 	}
@@ -771,20 +961,39 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 	}
 
 	/* these are ignored for SMB2 */
-	in_create_options &= ~(0x10);/* NTCREATEX_OPTIONS_SYNC_ALERT */
-	in_create_options &= ~(0x20);/* NTCREATEX_OPTIONS_ASYNC_ALERT */
+	state->in_create_options &= ~(0x10); /* NTCREATEX_OPTIONS_SYNC_ALERT */
+	state->in_create_options &= ~(0x20); /* NTCREATEX_OPTIONS_ASYNC_ALERT */
 
 	in_file_attributes &= ~FILE_FLAG_POSIX_SEMANTICS;
 
 	is_dfs = (smb1req->flags2 & FLAGS2_DFS_PATHNAMES);
 	if (is_dfs) {
+		const char *non_dfs_in_name = NULL;
 		/*
-		 * With a DFS flag set, remove any leading '\\'
-		 * characters from in_name before further processing.
+		 * With a DFS flag set, remove any DFS prefix
+		 * before further processing.
 		 */
-		while (in_name[0] == '\\') {
-			in_name++;
+		status = smb2_strip_dfs_path(in_name, &non_dfs_in_name);
+		if (!NT_STATUS_IS_OK(status)) {
+			tevent_req_nterror(req, status);
+			return tevent_req_post(req, state->ev);
 		}
+		/*
+		 * TODO: Note for dealing with reparse point errors.
+		 * We will need to remember and store the number of characters
+		 * we have removed here, which is (non_dfs_in_name - in_name)
+		 * in order to correctly report how many characters we
+		 * have removed before hitting the reparse point.
+		 * This will be a patch needed once we properly
+		 * deal with reparse points later.
+		 */
+		in_name = non_dfs_in_name;
+		/*
+		 * Now we're no longer dealing with a DFS path, so
+		 * remove the flag.
+		 */
+		smb1req->flags2 &= ~FLAGS2_DFS_PATHNAMES;
+		is_dfs = false;
 	}
 
 	state->fname = talloc_strdup(state, in_name);
@@ -851,7 +1060,8 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 	}
 
 	/* Check for trailing slash specific directory handling. */
-	status = windows_name_trailing_check(state->fname, in_create_options);
+	status = windows_name_trailing_check(state->fname,
+					     state->in_create_options);
 	if (tevent_req_nterror(req, status)) {
 		return tevent_req_post(req, state->ev);
 	}
@@ -869,29 +1079,42 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 	 * there is nothing else to do), durable_reconnect or
 	 * new open.
 	 */
-	if (state->replay_operation) {
+	if (state->replay_operation && !state->replay_reconnect) {
+		SMB_ASSERT(state->op != NULL);
 		state->result = state->op->compat;
 		state->result->op = state->op;
 		state->update_open = false;
-		state->info = state->op->create_action;
+		state->info = state->op->global->create_action;
 
 		smbd_smb2_create_after_exec(req);
 		if (!tevent_req_is_in_progress(req)) {
-			return req;
+			return tevent_req_post(req, state->ev);
 		}
 
 		smbd_smb2_create_finish(req);
 		return req;
 	}
 
-	if (state->do_durable_reconnect) {
+	if (state->do_durable_reconnect || state->replay_reconnect) {
 		DATA_BLOB new_cookie = data_blob_null;
 		NTTIME now = timeval_to_nttime(&smb2req->request_time);
+		const struct smb2_lease_key *lease_key = NULL;
 
+		/*
+		 * Assert a replay on a multichannel connection doesn't end up
+		 * here.
+		 */
+		SMB_ASSERT(state->op == NULL);
+
+		if (state->lease_ptr != NULL) {
+			lease_key = &state->lease_ptr->lease_key;
+		}
 		status = smb2srv_open_recreate(smb2req->xconn,
-					       smb1req->conn->session_info,
+					       smb2req->session,
+					       smb2req->tcon,
 					       state->persistent_id,
 					       state->create_guid,
+					       lease_key,
 					       now,
 					       &state->op);
 		if (tevent_req_nterror(req, status)) {
@@ -926,6 +1149,7 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 				   nt_errstr(status),
 				   nt_errstr(return_status));
 
+			TALLOC_FREE(state->op);
 			tevent_req_nterror(req, return_status);
 			return tevent_req_post(req, state->ev);
 		}
@@ -947,16 +1171,20 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 		state->op->status = NT_STATUS_OK;
 		state->op->global->disconnect_time = 0;
 
-		/* save the timout for later update */
+		/* save the timeout for later update */
 		state->durable_timeout_msec = state->op->global->durable_timeout_msec;
 
 		state->update_open = true;
 
-		state->info = FILE_WAS_OPENED;
+		if (!state->replay_reconnect) {
+			state->info = FILE_WAS_OPENED;
+		} else {
+			state->info = state->op->global->create_action;
+		}
 
 		smbd_smb2_create_after_exec(req);
 		if (!tevent_req_is_in_progress(req)) {
-			return req;
+			return tevent_req_post(req, state->ev);
 		}
 
 		smbd_smb2_create_finish(req);
@@ -971,23 +1199,54 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 		state->lease_ptr = NULL;
 	}
 
+	is_posix = (state->posx != NULL);
+
 	/* convert '\\' into '/' */
-	status = check_path_syntax_smb2(state->fname, is_dfs);
+	status = check_path_syntax(state->fname, is_posix);
 	if (tevent_req_nterror(req, status)) {
 		return tevent_req_post(req, state->ev);
 	}
 
-	ucf_flags = filename_create_ucf_flags(
-		smb1req, state->in_create_disposition);
+	ucf_flags = filename_create_ucf_flags(smb1req,
+					      state->in_create_disposition,
+					      state->in_create_options);
 
-	status = filename_convert_dirfsp(
-		req,
-		smb1req->conn,
-		state->fname,
-		ucf_flags,
-		state->twrp_time,
-		&dirfsp,
-		&smb_fname);
+	if (lp_follow_symlinks(SNUM(smb1req->conn)) &&
+	    (state->posx == NULL)) {
+		status = filename_convert_dirfsp(mem_ctx,
+						 smb1req->conn,
+						 state->fname,
+						 ucf_flags,
+						 state->twrp_time,
+						 &dirfsp,
+						 &smb_fname);
+	} else {
+		struct smb_filename *smb_fname_rel = NULL;
+
+		status = filename_convert_dirfsp_nosymlink(
+			mem_ctx,
+			smb1req->conn,
+			smb1req->conn->cwd_fsp,
+			state->fname,
+			ucf_flags,
+			state->twrp_time,
+			&dirfsp,
+			&smb_fname,
+			&smb_fname_rel,
+			&state->symlink_err);
+		TALLOC_FREE(smb_fname_rel);
+
+		if ((state->symlink_err != NULL) &&
+		    !(state->in_create_options & FILE_OPEN_REPARSE_POINT))
+		{
+			if (dirfsp != NULL) {
+				close_file_free(NULL, &dirfsp, ERROR_CLOSE);
+			}
+			TALLOC_FREE(smb_fname);
+			TALLOC_FREE(smb_fname_rel);
+			status = NT_STATUS_STOPPED_ON_SYMLINK;
+		}
+	}
 	if (tevent_req_nterror(req, status)) {
 		return tevent_req_post(req, state->ev);
 	}
@@ -1020,10 +1279,17 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 	 * server MUST fail the request with
 	 * STATUS_INVALID_PARAMETER.
 	 */
-	if (in_name[0] == '\\' || in_name[0] == '/') {
-		tevent_req_nterror(req,
-				   NT_STATUS_INVALID_PARAMETER);
-		return tevent_req_post(req, state->ev);
+	if (in_name[0] == '/') {
+		/* Names starting with '/' are never allowed. */
+		tevent_req_nterror(req, NT_STATUS_INVALID_PARAMETER);
+		return tevent_req_post(req, ev);
+	}
+	if (!is_posix && (in_name[0] == '\\')) {
+		/*
+		 * Windows names starting with '\' are not allowed.
+		 */
+		tevent_req_nterror(req, NT_STATUS_INVALID_PARAMETER);
+		return tevent_req_post(req, ev);
 	}
 
 	status = SMB_VFS_CREATE_FILE(smb1req->conn,
@@ -1033,7 +1299,7 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 				     in_desired_access,
 				     in_share_access,
 				     state->in_create_disposition,
-				     in_create_options,
+				     state->in_create_options,
 				     in_file_attributes,
 				     map_smb2_oplock_levels_to_samba(
 					     state->requested_oplock_level),
@@ -1046,9 +1312,25 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 				     &state->info,
 				     &in_context_blobs,
 				     state->out_context_blobs);
+	if (NT_STATUS_IS_OK(status) &&
+	    !(state->in_create_options & FILE_OPEN_REPARSE_POINT))
+	{
+
+		mode_t mode = state->result->fsp_name->st.st_ex_mode;
+
+		if (!(S_ISREG(mode) || S_ISDIR(mode))) {
+			/*
+			 * Only open files and dirs without
+			 * FILE_OPEN_REPARSE_POINT
+			 */
+			close_file_free(smb1req, &state->result, ERROR_CLOSE);
+			status = NT_STATUS_IO_REPARSE_TAG_NOT_HANDLED;
+		}
+	}
 	if (!NT_STATUS_IS_OK(status)) {
 		if (open_was_deferred(smb1req->xconn, smb1req->mid)) {
-			SMBPROFILE_IOBYTES_ASYNC_SET_IDLE(smb2req->profile);
+			SMBPROFILE_IOBYTES_ASYNC_SET_IDLE_X(smb2req->profile,
+							    smb2req->profile_x);
 			return req;
 		}
 		tevent_req_nterror(req, status);
@@ -1056,9 +1338,15 @@ static struct tevent_req *smbd_smb2_create_send(TALLOC_CTX *mem_ctx,
 	}
 	state->op = state->result->op;
 
+	if ((state->in_create_disposition == FILE_SUPERSEDE) &&
+	    (state->info == FILE_WAS_OVERWRITTEN))
+	{
+		state->info = FILE_WAS_SUPERSEDED;
+	}
+
 	smbd_smb2_create_after_exec(req);
 	if (!tevent_req_is_in_progress(req)) {
-		return req;
+		return tevent_req_post(req, state->ev);
 	}
 
 	smbd_smb2_create_finish(req);
@@ -1090,6 +1378,135 @@ static void smbd_smb2_create_purge_replay_cache(struct tevent_req *req,
 	state->purge_create_guid = NULL;
 }
 
+static void smbd_smb2_cc_before_exec_dhc2q(struct tevent_req *req)
+{
+	struct smbd_smb2_create_state *state = tevent_req_data(
+		req, struct smbd_smb2_create_state);
+	struct smbd_smb2_request *smb2req = state->smb2req;
+	const uint8_t *p = state->dh2q->data.data;
+	NTTIME now = timeval_to_nttime(&smb2req->request_time);
+	uint32_t durable_v2_timeout = 0;
+	DATA_BLOB create_guid_blob;
+	const uint8_t *hdr = NULL;
+	uint32_t flags;
+	NTSTATUS status;
+
+	if (state->dh2q->data.length != 32) {
+		tevent_req_nterror(req, NT_STATUS_INVALID_PARAMETER);
+		return;
+	}
+
+	if (state->dhnq != NULL) {
+		tevent_req_nterror(req, NT_STATUS_INVALID_PARAMETER);
+		return;
+	}
+
+	durable_v2_timeout = IVAL(p, 0);
+	create_guid_blob = data_blob_const(p + 16, 16);
+
+	status = GUID_from_ndr_blob(&create_guid_blob,
+				    &state->_create_guid);
+	if (tevent_req_nterror(req, status)) {
+		return;
+	}
+	state->create_guid = &state->_create_guid;
+
+	/*
+	 * we need to store the create_guid later
+	 */
+	state->update_open = true;
+
+	/*
+	 * And we need to create a cache for replaying the
+	 * create.
+	 */
+	state->need_replay_cache = true;
+
+	/*
+	 * durable handle v2 request processed below
+	 */
+	state->durable_requested = true;
+	state->durable_timeout_msec = MIN(durable_v2_timeout, 300*1000);
+	if (state->durable_timeout_msec == 0) {
+		/*
+		 * Set the timeout to 1 min as default.
+		 *
+		 * This matches Windows 2012.
+		 */
+		state->durable_timeout_msec = (60*1000);
+	}
+
+	/*
+	 * Check for replay operation.
+	 * Only consider it when we have dh2q.
+	 * If we do not have a replay operation, verify that
+	 * the create_guid is not cached for replay.
+	 */
+	hdr = SMBD_SMB2_IN_HDR_PTR(smb2req);
+	flags = IVAL(hdr, SMB2_HDR_FLAGS);
+	state->replay_operation = flags & SMB2_HDR_FLAG_REPLAY_OPERATION;
+
+	if (state->open_was_deferred) {
+		/*
+		 * When processing a redispatched deferred open, we have the
+		 * following state:
+		 * - no OPEN record
+		 * - RC record with global_file_id=0 (pending open state)
+		 *
+		 * We just skip calling smb2srv_open_lookup_replay_cache() as
+		 * - we already have a RC record
+		 * - it would fail with NT_STATUS_FILE_NOT_AVAILABLE which is
+		 *   not what we want in this "internal replay" case
+		 *
+		 * So we just set replay_operation to false and move on.
+		 */
+		state->replay_operation = false;
+		return;
+	}
+
+	status = smb2srv_open_lookup_replay_cache(smb2req->xconn,
+						  smb2req->session,
+						  *state->create_guid,
+						  state->fname,
+						  now,
+						  &state->persistent_id,
+						  &state->op);
+	if (NT_STATUS_EQUAL(status, NT_STATUS_FWP_RESERVED)) {
+		/*
+		 * We've reserved the replay_cache record
+		 * for ourself, indicating we're still
+		 * in progress.
+		 *
+		 * It means the smbd_smb2_create_cleanup()
+		 * may need to call smbXsrv_open_purge_replay_cache()
+		 * in order to cleanup.
+		 */
+		SMB_ASSERT(state->op == NULL);
+		state->_purge_create_guid = state->_create_guid;
+		state->purge_create_guid = &state->_purge_create_guid;
+		state->replay_operation = false;
+	} else if (NT_STATUS_EQUAL(status, NT_STATUS_FILE_NOT_AVAILABLE)) {
+		tevent_req_nterror(req, status);
+		return;
+	} else if (NT_STATUS_EQUAL(status, NT_STATUS_HANDLE_NO_LONGER_VALID)) {
+		state->replay_reconnect = true;
+	} else if (tevent_req_nterror(req, status)) {
+		DBG_WARNING("smb2srv_open_lookup_replay_cache "
+			    "failed: %s\n", nt_errstr(status));
+		return;
+	} else if (!state->replay_operation) {
+		/*
+		 * If a create without replay operation flag
+		 * is sent but with a create_guid that is
+		 * currently in the replay cache -- fail.
+		 */
+		(void)tevent_req_nterror(req, NT_STATUS_DUPLICATE_OBJECTID);
+		return;
+	}
+
+	return;
+}
+
 static void smbd_smb2_create_before_exec(struct tevent_req *req)
 {
 	struct smbd_smb2_create_state *state = tevent_req_data(
@@ -1113,13 +1530,8 @@ static void smbd_smb2_create_before_exec(struct tevent_req *req)
 			return;
 		}
 
-		/*
-		 * NB. When SMB2+ unix extensions are added,
-		 * we need to relax this check in invalid
-		 * names - we used to not do this if
-		 * lp_posix_pathnames() was false.
-		 */
-		if (ea_list_has_invalid_name(state->ea_list)) {
+		if ((state->posx == NULL) &&
+		    ea_list_has_invalid_name(state->ea_list)) {
 			tevent_req_nterror(req, STATUS_INVALID_EA_NAME);
 			return;
 		}
@@ -1180,105 +1592,8 @@ static void smbd_smb2_create_before_exec(struct tevent_req *req)
 	}
 
 	if (state->dh2q != NULL) {
-		const uint8_t *p = state->dh2q->data.data;
-		NTTIME now = timeval_to_nttime(&smb2req->request_time);
-		uint32_t durable_v2_timeout = 0;
-		DATA_BLOB create_guid_blob;
-		const uint8_t *hdr;
-		uint32_t flags;
-
-		if (state->dh2q->data.length != 32) {
-			tevent_req_nterror(req, NT_STATUS_INVALID_PARAMETER);
-			return;
-		}
-
-		if (state->dhnq != NULL) {
-			tevent_req_nterror(req, NT_STATUS_INVALID_PARAMETER);
-			return;
-		}
-
-		durable_v2_timeout = IVAL(p, 0);
-		create_guid_blob = data_blob_const(p + 16, 16);
-
-		status = GUID_from_ndr_blob(&create_guid_blob,
-					    &state->_create_guid);
-		if (tevent_req_nterror(req, status)) {
-			return;
-		}
-		state->create_guid = &state->_create_guid;
-
-		/*
-		 * we need to store the create_guid later
-		 */
-		state->update_open = true;
-
-		/*
-		 * And we need to create a cache for replaying the
-		 * create.
-		 */
-		state->need_replay_cache = true;
-
-		/*
-		 * durable handle v2 request processed below
-		 */
-		state->durable_requested = true;
-		state->durable_timeout_msec = MIN(durable_v2_timeout, 300*1000);
-		if (state->durable_timeout_msec == 0) {
-			/*
-			 * Set the timeout to 1 min as default.
-			 *
-			 * This matches Windows 2012.
-			 */
-			state->durable_timeout_msec = (60*1000);
-		}
-
-		/*
-		 * Check for replay operation.
-		 * Only consider it when we have dh2q.
-		 * If we do not have a replay operation, verify that
-		 * the create_guid is not cached for replay.
-		 */
-		hdr = SMBD_SMB2_IN_HDR_PTR(smb2req);
-		flags = IVAL(hdr, SMB2_HDR_FLAGS);
-		state->replay_operation =
-			flags & SMB2_HDR_FLAG_REPLAY_OPERATION;
-
-		status = smb2srv_open_lookup_replay_cache(smb2req->xconn,
-							  state->req_guid,
-							  *state->create_guid,
-							  state->fname,
-							  now,
-							  &state->op);
-		if (NT_STATUS_EQUAL(status, NT_STATUS_FWP_RESERVED)) {
-			/*
-			 * We've reserved the replay_cache record
-			 * for ourself, indicating we're still
-			 * in progress.
-			 *
-			 * It means the smbd_smb2_create_cleanup()
-			 * may need to call smbXsrv_open_purge_replay_cache()
-			 * in order to cleanup.
-			 */
-			SMB_ASSERT(state->op == NULL);
-			state->_purge_create_guid = state->_create_guid;
-			state->purge_create_guid = &state->_purge_create_guid;
-			status = NT_STATUS_OK;
-			state->replay_operation = false;
-		} else if (NT_STATUS_EQUAL(status, NT_STATUS_FILE_NOT_AVAILABLE)) {
-			tevent_req_nterror(req, status);
-			return;
-		} else if (tevent_req_nterror(req, status)) {
-			DBG_WARNING("smb2srv_open_lookup_replay_cache "
-				    "failed: %s\n", nt_errstr(status));
-			return;
-		} else if (!state->replay_operation) {
-			/*
-			 * If a create without replay operation flag
-			 * is sent but with a create_guid that is
-			 * currently in the replay cache -- fail.
-			 */
-			status = NT_STATUS_DUPLICATE_OBJECTID;
-			(void)tevent_req_nterror(req, status);
+		smbd_smb2_cc_before_exec_dhc2q(req);
+		if (!tevent_req_is_in_progress(req)) {
 			return;
 		}
 	}
@@ -1365,7 +1680,7 @@ static void smbd_smb2_create_before_exec(struct tevent_req *req)
 		 * established open carries a lease with the
 		 * same lease key.
 		 */
-		if (state->replay_operation) {
+		if (state->replay_operation && !state->replay_reconnect) {
 			struct smb2_lease *op_ls =
 				&state->op->compat->lease->lease;
 			int op_oplock = state->op->compat->oplock_type;
@@ -1408,11 +1723,63 @@ static void smbd_smb2_create_after_exec(struct tevent_req *req)
 	 * here we have op == result->op
 	 */
 
-	DEBUG(10, ("smbd_smb2_create_send: "
-		   "response construction phase\n"));
+	DBG_DEBUG("response construction phase\n");
+
+	state->op->global->create_action = state->info;
+	state->out_create_action = state->info;
 
 	state->out_file_attributes = fdos_mode(state->result);
 
+	if ((state->out_file_attributes & FILE_ATTRIBUTE_REPARSE_POINT) &&
+	    (!(state->in_create_options & FILE_OPEN_REPARSE_POINT)))
+	{
+
+		uint32_t tag;
+		uint8_t *data = NULL;
+		uint32_t len = 0;
+
+		status = fsctl_get_reparse_point(
+			state->result, talloc_tos(), &tag, &data, 65536, &len);
+		if (!NT_STATUS_IS_OK(status)) {
+			goto fail;
+		}
+
+		if (tag != IO_REPARSE_TAG_SYMLINK) {
+			status = NT_STATUS_IO_REPARSE_TAG_NOT_HANDLED;
+			goto fail;
+		}
+
+		state->symlink_err = talloc_zero(state,
+						 struct reparse_data_buffer);
+		if (state->symlink_err == NULL) {
+			TALLOC_FREE(data);
+			status = NT_STATUS_NO_MEMORY;
+			goto fail;
+		}
+
+		status = reparse_data_buffer_parse(state->symlink_err,
+						   state->symlink_err,
+						   data,
+						   len);
+		TALLOC_FREE(data);
+		if (!NT_STATUS_IS_OK(status)) {
+			DBG_DEBUG("reparse_data_buffer_parse failed: %s\n",
+				  nt_errstr(status));
+			goto fail;
+		}
+
+		/*
+		 * Checked above, just to make sure
+		 */
+		SMB_ASSERT(state->symlink_err->tag == IO_REPARSE_TAG_SYMLINK);
+
+		DBG_DEBUG("Redirecting to %s\n",
+			  state->symlink_err->parsed.lnk.substitute_name);
+
+		status = NT_STATUS_STOPPED_ON_SYMLINK;
+		goto fail;
+	}
+
 	if (state->mxac != NULL) {
 		NTTIME last_write_time;
 
@@ -1578,46 +1945,42 @@ static void smbd_smb2_create_after_exec(struct tevent_req *req)
 	}
 
 	if (state->posx != NULL) {
-		struct dom_sid owner = { .sid_rev_num = 0, };
-		struct dom_sid group = { .sid_rev_num = 0, };
 		struct stat_ex *psbuf = &state->result->fsp_name->st;
-		ssize_t cc_len;
+		struct smb3_posix_cc_info cc = {
+			.nlinks = psbuf->st_ex_nlink,
+			.posix_mode = unix_mode_to_wire(psbuf->st_ex_mode),
+		};
+		uint8_t buf[sizeof(struct smb3_posix_cc_info)];
+		struct ndr_push ndr = {
+			.data = buf,
+			.alloc_size = sizeof(buf),
+			.fixed_buf_size = true,
+		};
+		enum ndr_err_code ndr_err;
 
-		uid_to_sid(&owner, psbuf->st_ex_uid);
-		gid_to_sid(&group, psbuf->st_ex_gid);
+		uid_to_sid(&cc.owner, psbuf->st_ex_uid);
+		gid_to_sid(&cc.group, psbuf->st_ex_gid);
 
-		cc_len = smb2_posix_cc_info(
-			conn, 0, psbuf, &owner, &group, NULL, 0);
+		(void)fsctl_get_reparse_tag(state->result, &cc.reparse_tag);
 
-		if (cc_len == -1) {
+		ndr_err =
+			ndr_push_smb3_posix_cc_info(&ndr,
+						    NDR_SCALARS | NDR_BUFFERS,
+						    &cc);
+		if (!NDR_ERR_CODE_IS_SUCCESS(ndr_err)) {
 			status = NT_STATUS_INSUFFICIENT_RESOURCES;
 			goto fail;
 		}
 
-		{
-			/*
-			 * cc_len is 68 + 2 SIDs, allocate on the stack
-			 */
-			uint8_t buf[cc_len];
-			DATA_BLOB blob = { .data = buf, .length = cc_len, };
-
-			smb2_posix_cc_info(
-				conn,
-				0,
-				psbuf,
-				&owner,
-				&group,
-				buf,
-				sizeof(buf));
-
-			status = smb2_create_blob_add(
-				state->out_context_blobs,
-				state->out_context_blobs,
-				SMB2_CREATE_TAG_POSIX,
-				blob);
-			if (!NT_STATUS_IS_OK(status)) {
-				goto fail;
-			}
+		status = smb2_create_blob_add(state->out_context_blobs,
+					      state->out_context_blobs,
+					      SMB2_CREATE_TAG_POSIX,
+					      (DATA_BLOB){
+						      .data = buf,
+						      .length = ndr.offset,
+					      });
+		if (!NT_STATUS_IS_OK(status)) {
+			goto fail;
 		}
 	}
 
@@ -1626,7 +1989,6 @@ static void smbd_smb2_create_after_exec(struct tevent_req *req)
 fail:
 	close_file_free(state->smb1req, &state->result, ERROR_CLOSE);
 	tevent_req_nterror(req, status);
-	tevent_req_post(req, state->ev);
 }
 
 static void smbd_smb2_create_finish(struct tevent_req *req)
@@ -1647,20 +2009,11 @@ static void smbd_smb2_create_finish(struct tevent_req *req)
 		state->out_oplock_level	= map_samba_oplock_levels_to_smb2(result->oplock_type);
 	}
 
-	if ((state->in_create_disposition == FILE_SUPERSEDE)
-	    && (state->info == FILE_WAS_OVERWRITTEN)) {
-		state->out_create_action = FILE_WAS_SUPERSEDED;
-	} else {
-		state->out_create_action = state->info;
-	}
-	result->op->create_action = state->out_create_action;
-
 	state->out_creation_ts = get_create_timespec(smb1req->conn,
 					result, result->fsp_name);
 	state->out_last_access_ts = result->fsp_name->st.st_ex_atime;
 	state->out_last_write_ts = result->fsp_name->st.st_ex_mtime;
-	state->out_change_ts = get_change_timespec(smb1req->conn,
-					result, result->fsp_name);
+	state->out_change_ts = result->fsp_name->st.st_ex_ctime;
 
 	if (lp_dos_filetime_resolution(SNUM(smb2req->tcon->compat))) {
 		dos_filetime_timespec(&state->out_creation_ts);
@@ -1698,13 +2051,60 @@ static NTSTATUS smbd_smb2_create_recv(struct tevent_req *req,
 			uint32_t *out_file_attributes,
 			uint64_t *out_file_id_persistent,
 			uint64_t *out_file_id_volatile,
-			struct smb2_create_blobs *out_context_blobs)
+			struct smb2_create_blobs *out_context_blobs,
+			struct reparse_data_buffer **symlink_reparse)
 {
-	NTSTATUS status;
+	NTSTATUS status = NT_STATUS_OK;
 	struct smbd_smb2_create_state *state = tevent_req_data(req,
 					       struct smbd_smb2_create_state);
+	bool error;
 
-	if (tevent_req_is_nterror(req, &status)) {
+	error = tevent_req_is_nterror(req, &status);
+
+	if (NT_STATUS_EQUAL(status, NT_STATUS_STOPPED_ON_SYMLINK)) {
+		struct symlink_reparse_struct *lnk = &state->symlink_err
+							      ->parsed.lnk;
+		size_t fname_len = strlen(state->fname);
+
+		/*
+		 * filename_convert_dirfsp_nosymlink() calculates
+		 * unparsed_path_length. Just assert it did not mess up big
+		 * time.
+		 */
+		SMB_ASSERT(lnk->unparsed_path_length <= fname_len);
+
+		if (lnk->unparsed_path_length != 0) {
+			bool ok;
+			char *unparsed_unix = NULL;
+			char *utf_16 = NULL;
+			size_t utf_16_len;
+
+			unparsed_unix = state->fname + fname_len -
+					lnk->unparsed_path_length;
+
+			ok = convert_string_talloc(talloc_tos(),
+						   CH_UNIX,
+						   CH_UTF16,
+						   unparsed_unix,
+						   lnk->unparsed_path_length,
+						   &utf_16,
+						   &utf_16_len);
+			if (!ok) {
+				tevent_req_received(req);
+				return NT_STATUS_INTERNAL_ERROR;
+			}
+			TALLOC_FREE(utf_16);
+
+			if (utf_16_len > UINT16_MAX) {
+				tevent_req_received(req);
+				return NT_STATUS_BUFFER_OVERFLOW;
+			}
+			lnk->unparsed_path_length = utf_16_len;
+		}
+		*symlink_reparse = talloc_move(mem_ctx, &state->symlink_err);
+	}
+
+	if (error) {
 		tevent_req_received(req);
 		return status;
 	}
@@ -1721,6 +2121,7 @@ static NTSTATUS smbd_smb2_create_recv(struct tevent_req *req,
 	*out_file_id_persistent	= state->out_file_id_persistent;
 	*out_file_id_volatile	= state->out_file_id_volatile;
 	*out_context_blobs	= *(state->out_context_blobs);
+	*symlink_reparse	= NULL;
 
 	talloc_steal(mem_ctx, state->out_context_blobs->blobs);
 
